from ctREFPROP.ctREFPROP import REFPROPFunctionLibrary
import REFPROPConnector.constants as constants
from abc import ABC, abstractmethod
import warnings


GLOBALCounter = 0


class RefPropHandler:

    def __init__(self, fluids: list, composition: list, unit_system="SI WITH C"):

        # If REFPROP is available the program uses it, otherwise COOLPROP is used

        #   DEFAULT MEASURE UNITS:
        #
        #   Temperature:        [°C]
        #   Pressure:           [MPa]
        #   Density:            [kg/m^3]
        #   Enthalpy:           [kJ/kg]
        #   Entropy:            [kJ/(kg*K)]
        #   Speed:              [m/s]
        #   Kinematic vis.:     [cm^2/s]
        #   Viscosity:          [uPa*s]
        #   Thermal cond.:      [mW/(m*K)]
        #   Surface tension:    [mN/m]
        #   Molar Mass:         [kg/kmol]
        #   Heat Capacity:      [kJ/(kg*K)]

        self.refprop = REFPROPFunctionLibrary(constants.RP_EXEC)
        self.refprop.SETPATHdll(constants.RP_EXEC)

        self.fluids = fluids
        self.composition = composition
        self.unit_system = unit_system

        self.TC = self.calculate("", "TC", 0, 0)
        self.PC = self.calculate("", "PC", 0, 0)

    def __initialize_reference_state(self):

        T_unit = self.return_units("T")
        P_unit = self.return_units("P")

        T_0 = 20        # °C
        P_0 = 101325    # Pa

        if T_unit == "C":

            # Celsius
            self.T_0 = T_0

        elif T_unit == "K":

            # Kelvin
            self.T_0 = T_0 + 273.15

        else:

            # Fahrenheit
            self.T_0 = T_0 * 9 / 5 + 32

        if P_unit == "Pa":

            # Pascal
            self.P_0 = P_0

        elif P_unit == "kPa":

            # kiloPascal
            self.P_0 = P_0 / 1000

        elif P_unit == "bar":

            # bar
            self.P_0 = P_0 / 100000

        elif P_unit == "MPa":

            # bar
            self.P_0 = P_0 / 100000

        else:

            # PSI
            self.P_0 = P_0 / 6894.7572931783

        self.H_0 = self.calculate("TP", "H", self.T_0, self.P_0)
        self.S_0 = self.calculate("TP", "s", self.T_0, self.P_0)

    def calculate(self, str_in: str, str_out: str, a: float, b: float):

        try:
            self.refprop.SETFLUIDSdll('*'.join(self.fluids))
            return self.refprop.REFPROP1dll(str_in, str_out, self.SI, 1, a, b, self.composition).c
        except:

            a= 20

    def return_units(self, property_name):

        return_value = constants.get_units(property_name, self.unit_system)

        if "Unknown" in return_value:

            return ""

        else:

            return return_value

    @property
    def unit_system(self):

        return self.__unit_system

    @unit_system.setter
    def unit_system(self, unit_system_input):

        self.__unit_system = unit_system_input

        try:

            self.SI = self.refprop.GETENUMdll(0, self.__unit_system).iEnum

        except:

            self.SI = self.refprop.GETENUMdll(0, "SI WITH C").iEnum

            warning_message = (

                "{} unit system is not supported, "
                "{} has been used instead\n"
                "Check in the refprop manual for the correct"
                "name of the system that you wanted to use"

            ).format(self.__unit_system, "SI WITH C")

            warnings.warn(warning_message)
            self.__unit_system = "SI WITH C"

        self.__initialize_reference_state()


class ThermodynamicVariable:

    def __init__(self, name: str):

        self.value = None
        self.name = name
        self.refprop_name = constants.get_refprop_name(name)
        self.is_user_defined = False
        self.order = 0

    @property
    def is_empty(self):
        return self.value is None

    def __gt__(self, other):
        # enables comparison
        # self > other

        return self.order > other.order

    def __lt__(self, other):
        # enables comparison
        # self < other

        return self.order < other.order

    def __le__(self, other):
        return not self.__gt__(other)

    def __ge__(self, other):
        return not self.__lt__(other)


class AbstractThermodynamicPoint(ABC):

    @classmethod
    def init_from_fluid(cls, fluids: list, composition: list, other_variables="all", calculate_on_need="all", unit_system="SI WITH C"):

        RP = RefPropHandler(fluids, composition, unit_system)
        return cls(RP, other_variables, calculate_on_need)

    def __init__(self, refprop: RefPropHandler, other_variables="all", calculate_on_need="all"):

        self.RPHandler = refprop

        self.__initialize_state_variables()
        self.__initialize_other_variables(other_variables)
        self.reset_state_var_order()

        self.calculated_variables = list()
        self.__initialize_calculate_on_need_variables(calculate_on_need)

    def __initialize_state_variables(self):

        self.state_var_list = list()

        for variable_name in ["P", "T", "h", "s", "Q", "rho"]:

            new_variable = ThermodynamicVariable(variable_name)
            self.state_var_list.append(new_variable)

    def __initialize_other_variables(self, other_variables):

        self.other_variables = list()

        if other_variables is not None:

            if other_variables == "all":

                other_variables = constants.get_all_refprop_names()

            if type(other_variables) == list:

                state_variables_refprop_names = self.state_variables_refprop_names

                for variable_name in other_variables:

                    if not constants.get_refprop_name(variable_name) in state_variables_refprop_names:

                        new_variable = ThermodynamicVariable(variable_name)
                        self.other_variables.append(new_variable)

    def __initialize_calculate_on_need_variables(self, calculate_on_need):

        self.calculate_on_need_variables = list()

        if calculate_on_need is not None:

            if type(calculate_on_need) == list:

                for varible_name in calculate_on_need:

                    found_variable = self.__get_variable_from_name(varible_name)

                    if found_variable is not None:
                        self.calculate_on_need_variables.append(found_variable)

            elif calculate_on_need == "all":

                self.calculate_on_need_variables.extend(self.variables)

            elif calculate_on_need == "other":

                self.calculate_on_need_variables.extend(self.other_variables)

            elif calculate_on_need == "state":

                self.calculate_on_need_variables.extend(self.state_var_list)

    def __calculate(self):

        not_none_variables = self.not_none_variables
        self.calculated_variables = not_none_variables

        for variable in self.state_var_list:

            if variable in not_none_variables:

                variable.is_user_defined = True

            else:

                if variable not in self.calculate_on_need_variables:
                    self.__calculate_variable(variable)

        for variable in self.other_variables:

            if variable not in self.calculate_on_need_variables:
                self.__calculate_variable(variable)

        self.other_calculation()

    def __calculate_variable(self, variable):

        global GLOBALCounter

        if self.calculation_ready:

            GLOBALCounter += 1

            not_none_variables = self.not_none_variables
            input_str = not_none_variables[0].refprop_name + not_none_variables[1].refprop_name

            variable.value = self.RPHandler.calculate(input_str,
                                                      variable.refprop_name,
                                                      not_none_variables[0].value,
                                                      not_none_variables[1].value)

            self.calculated_variables.append(variable)

    @abstractmethod
    def other_calculation(self):
        pass

    def set_variable(self, variable_name: str, variable_value: float):

        input_refprop_name = constants.get_refprop_name(variable_name)

        for variable in self.state_var_list:

            if variable.refprop_name == input_refprop_name:
                variable.value = variable_value
                variable.order = 0
                break

        self.state_var_list.sort()
        self.reset_state_var_order()

        if self.calculation_ready:
            self.__calculate()

    def get_variable(self, variable_name: str):

        variable = self.__get_variable_from_name(variable_name)

        if variable is not None:

            if variable not in self.calculated_variables:
                self.__calculate_variable(variable)

            return variable.value

        if variable_name == "exergy":

            h = self.get_variable("h")
            s = self.get_variable("s")

            return (h - self.RPHandler.H_0) - self.RPHandler.T_0 * (s - self.RPHandler.S_0)

        try:

            return self.__dict__[variable_name]

        except:

            return None

    def get_unit(self, variable_name: str):

        return self.RPHandler.return_units(variable_name)

    def __get_variable_from_name(self, variable_name: str):

        input_refprop_name = constants.get_refprop_name(variable_name.lower())

        for variable in self.variables:

            if variable.refprop_name == input_refprop_name:
                return variable

        return None

    def reset_state_var_order(self):

        counter = 1
        for variable in self.state_var_list:
            variable.order = counter
            counter += 1

    @property
    def calculation_ready(self):

        counter = 0
        for variable in self.state_var_list:

            if not variable.is_empty:
                counter += 1

        if counter > 2:

            return True

        elif counter == 2:

            __Q = self.__get_variable_from_name("Q")

            if not __Q.is_empty:

                if (__Q.value <= 1) and (__Q.value >= 0):
                    return True

                else:
                    return False

            return True

        else:

            return False

    @property
    def not_none_variables(self) -> list:

        return_list = list()

        for variable in self.state_var_list:

            if not variable.is_empty:

                return_list.append(variable)

                if len(return_list) == 2:
                    break

        return return_list

    @property
    def state_variables_refprop_names(self)->list:

        return_list = list()
        for variable in self.state_var_list:

            return_list.append(variable.refprop_name)

        return return_list

    @property
    def variables(self):

        variables = list()
        variables.extend(self.state_var_list)
        variables.extend(self.other_variables)

        return variables

    @property
    def composition(self):

        return self.RPHandler.composition

    @composition.setter
    def composition(self, new_composition: list):

        self.RPHandler.composition = new_composition

    @staticmethod
    def print_global_counter(reset=True):

        global GLOBALCounter
        print(GLOBALCounter)

        if reset:
            GLOBALCounter = 0

    def list_properties(self):

        string_to_display = """
        
        -----------------------------------------------------------
        -----------------------------------------------------------
            
                        REFPROP_Connector VARIABLES
                
                (if you need more of them to be implemented
                please contact pietro.ungar@unifi.it)
        
        -----------------------------------------------------------
        -----------------------------------------------------------"""

        string_to_display += """
        
            STATE VARIABLES
            (variables that can be set in order to define the 
            thermodynamic state)
            """

        string_to_display += self.__return_variable_list_str(self.state_var_list)

        string_to_display += """
        -----------------------------------------------------------
        
            OTHER VARIABLES
            """

        string_to_display += self.__return_variable_list_str(self.other_variables)

        print(string_to_display)

    def list_unit_systems(self, step=4):

        i = 0

        string_to_display = """

                -----------------------------------------------------------
                -----------------------------------------------------------

                              REFPROP_Connector UNIT SYSTEMS

                        (if you need more of them to be implemented
                        please contact pietro.ungar@unifi.it)"""

        while i < len(self.variables):

            string_to_display += """

                -----------------------------------------------------------
                -----------------------------------------------------------

            """

            if  i + step < len(self.variables):

                string_to_display += self.__return_variable_unit_str(self.variables[i: i + step])

            else:

                string_to_display += self.__return_variable_unit_str(self.variables[i:])

            i = i + step

        print(string_to_display)

    def __return_variable_list_str(self, variable_list):

        string_to_display = """
            {:<15} {:<25} {:<10}
            """.format("REFPROP NAME:", "ALTERNATIVE NAMES:", "UNIT:")

        for variable in variable_list:

            refprop_name = variable.refprop_name
            other_std_names = constants.get_other_standard_names(refprop_name)
            unit = self.get_unit(refprop_name)

            string_to_display += """
            {:<15} {:<25}  {:<10}
            """.format(refprop_name.upper(), other_std_names[0], unit)

            string_to_display += """
            """.join("{:<15} {:<25}".format("", std_name) for std_name in other_std_names[1:])
            string_to_display += "\n"

        return string_to_display

    def __return_variable_unit_str(self, variable_list):

        BOLD = "\033[1m"
        RED = '\033[91m'
        END = "\033[0m"

        name_format = " {:15s}"
        name_bold = BOLD + name_format + END
        name_bold_red = RED + BOLD + name_format + END

        variable_format = " {:20s}"
        variable_bold = BOLD + variable_format + END
        variable_red = RED + variable_format + END

        string_to_display = "\n" + "{:<15} ".format("") + " ".join(

            variable_bold.format(variable.refprop_name.upper()) for variable in variable_list

        ) + "\n\n"

        for unit_system in constants.get_all_unit_systems():

            if unit_system == self.RPHandler.unit_system:

                string_to_display += name_bold_red.format(unit_system)
                string_to_display += " ".join(

                    variable_red.format(constants.get_units(variable.refprop_name, unit_system)) for variable in variable_list

                )

            else:

                string_to_display += name_bold.format(unit_system)
                string_to_display += " ".join(

                    variable_format.format(constants.get_units(variable.refprop_name, unit_system)) for variable in
                    variable_list

                )

            string_to_display += "\n"

        return string_to_display


class ThermodynamicPoint(AbstractThermodynamicPoint):

    def __init__(self, fluids: list, composition: list, other_variables="all", calculate_on_need="all", unit_system="SI WITH C"):

        RP = RefPropHandler(fluids, composition, unit_system)
        super().__init__(RP, other_variables=other_variables, calculate_on_need=calculate_on_need)

    def other_calculation(self):
        pass

    def init_from_fluid(cls, fluids: list, composition: list, other_variables="all", calculate_on_need="all", unit_system="SI WITH C"):

        return ThermodynamicPoint(fluids, composition, other_variables, calculate_on_need, unit_system)



from ctREFPROP.ctREFPROP import REFPROPFunctionLibrary
import xml.etree.ElementTree as ETree
from tkinter import filedialog as fd
import pyrebase.pyrebase as pyrebase
from abc import ABC, abstractmethod
import os, requests
import tkinter as tk

CURRENT_DIR = os.path.dirname(__file__)
REFPROP_NAMES_FILE = 'REFPROP_names.xml'
REFPROP_EXECUTABLE_PATH_FILE = 'REFPROP_exec.dat'
__REFPROP_PATH_FILE = os.path.join(CURRENT_DIR, REFPROP_EXECUTABLE_PATH_FILE)


# ---------------------------------
# ---------------------------------

# REFPROP EXECUTABLE IDENTIFICATION

# ---------------------------------
# ---------------------------------


def retreive_RP_exec(failure_possible=True):

    try:

        __RP_PATH = os.environ['RPPREFIX']

    except:

        root = tk.Tk()
        root.withdraw()
        __RP_PATH = fd.askdirectory(

            title='select REFPROP directory'

        )
        root.destroy()

    if os.path.isdir(__RP_PATH):

        try:

            REFPROPFunctionLibrary(__RP_PATH)

        except:

            raise FileNotFoundError(

                ("\n\nthe provided REFPROP executable does not work\n\n"
                 "\tExecutable provided: {}\n\n"
                 "Execution will stop!\n\n").format(__RP_PATH)

            )

        else:

            with open(__REFPROP_PATH_FILE, "w") as f:
                f.write(__RP_PATH)

            return __RP_PATH

    if not failure_possible:

        raise FileNotFoundError(

            ("\n\nREFPROP executable path must be provided!\n\n"
             "\t{} is not a suitable path\n\n"
             "Execution will stop!\n\n").format(__RP_PATH)

        )

    return None


RP_EXEC = None

if os.path.isfile(__REFPROP_PATH_FILE):

    if os.path.isfile(__REFPROP_PATH_FILE):

        with open(__REFPROP_PATH_FILE) as f:
            lines = f.readlines()

        if os.path.isfile(str(lines[0])):
            RP_EXEC = str(lines[0])

if RP_EXEC is None:

    RP_EXEC = retreive_RP_exec(failure_possible=False)


# ----------------------------------
# ----------------------------------

#   REFPROP NAMES TREE GENERATION

# ----------------------------------
# ----------------------------------

__FIREBASE_CONFIG = {

    "apiKey": "AIzaSyAWsQRNINNXzG87y4V9esY5NlcE9wbvdHY",
    "databaseURL": "https://serg-group-repository-default-rtdb.europe-west1.firebasedatabase.app",
    "authDomain": "serg-group-repository.firebaseapp.com",
    "projectId": "serg-group-repository",
    "storageBucket": "serg-group-repository.appspot.com",
    "messagingSenderId": "619279827614",
    "appId": "1:619279827614:web:563caa97267504b223718f",
    "measurementId": "G-007NTLM8RC"

}
__FIREBASE_TOKENS = {

    REFPROP_NAMES_FILE: "72f1c94d-ebdd-4b38-a887-ae048bf59071"

}


def _import_refprop_xml_files():
    for file_name in [REFPROP_NAMES_FILE]:

        file_path = os.path.join(CURRENT_DIR, file_name)

        if not os.path.isfile(file_path):

            try:

                firebase = pyrebase.initialize_app(__FIREBASE_CONFIG)
                storage = firebase.storage()
                storage.child("BHEModel/REFPROP_names.xml").download("", file_path, __FIREBASE_TOKENS[file_name])

                if not os.path.isfile(file_path):

                    url = storage.child("BHEModel/REFPROP_names.xml").get_url(__FIREBASE_TOKENS[file_name])
                    headers = {"Authorization": "Firebase " + __FIREBASE_TOKENS[file_name]}
                    r = requests.get(url, stream=True, headers=headers)

                    if r.status_code == 200:

                        return ETree.fromstring(r.content.decode("utf-8"))

                    else:

                        raise Exception("Unable to reach firebase storage, check your internet connection")

            except:

                raise Exception("Unable to reach firebase storage, check your internet connection")


def _get_refprop_name_xml(file_name) -> ETree.Element:
    _import_refprop_xml_files()

    if os.path.isfile(os.path.join(CURRENT_DIR, file_name)):
        tree = ETree.parse(os.path.join(CURRENT_DIR, file_name))
        return tree.getroot()


class __AbstractTree(ABC):

    def __init__(self):

        self.std_name = None
        self.left_tree = None
        self.right_tree = None

    def append_value(self, input_std_name, value):

        if self.is_empty:

            self.std_name = input_std_name
            self.init_value(value)

            self.left_tree = self.init_empty_self()
            self.right_tree = self.init_empty_self()

        else:

            if self.std_name < input_std_name:

                self.right_tree.append_value(input_std_name, value)

            elif self.std_name > input_std_name:

                self.left_tree.append_value(input_std_name, value)

    def find_element(self, input_std_name):

        if self.is_empty:

            return None

        else:

            if self.std_name == input_std_name:

                return self

            elif self.std_name < input_std_name:

                return self.right_tree.find_element(input_std_name)

            else:

                return self.left_tree.find_element(input_std_name)

    @property
    def is_empty(self):

        return self.std_name is None

    @abstractmethod
    def init_value(self, value):
        pass

    @classmethod
    @abstractmethod
    def init_empty_self(self):
        pass


class __RefPropNamesTree(__AbstractTree):

    def __init__(self):

        super().__init__()
        self.refprop_name = None
        self.units_dict = None

    def init_value(self, value):

        self.refprop_name = value["RP_Name"]
        self.units_dict = value["units_dict"]

    @classmethod
    def initialize_from_xml(cls):

        refprop_names_tree = cls()
        root = _get_refprop_name_xml(REFPROP_NAMES_FILE)

        for element in root.findall("refprop_name"):

            ref_prop_name = element.attrib["name"].lower()

            unit_dict = dict()

            for unit in element.find("units").findall("unit"):
                unit_dict.update({unit.attrib["name"]: unit.attrib["unit"]})

            value_dict = {"RP_Name": ref_prop_name, "units_dict": unit_dict}

            refprop_names_tree.append_value(ref_prop_name, value_dict)

            for std_name in element.find("std_names").findall("std_name"):
                refprop_names_tree.append_value(std_name.attrib["name"].lower(), value_dict)

        return refprop_names_tree

    @classmethod
    def init_empty_self(cls):

        return cls()

    def get_refprop_name(self, input_std_name):

        element_found = self.find_element(input_std_name)

        if element_found is not None:

            return element_found.refprop_name

        return None

    def get_units_dict(self, input_std_name):

        element_found = self.find_element(input_std_name)

        if element_found is not None:
            return element_found.units_dict

        return None

    def append_refprop_names(self, initial_list: list)->list:

        if self.is_empty:

            return initial_list

        else:

            new_list = self.right_tree.append_refprop_names(initial_list)
            new_list = self.left_tree.append_refprop_names(new_list)

            if self.refprop_name not in initial_list:

                new_list.append(self.refprop_name)

            return new_list

    def append_other_standard_names(self, initial_list: list, refprop_name)->list:

        if self.is_empty:

            return initial_list

        else:

            new_list = self.right_tree.append_other_standard_names(initial_list, refprop_name)
            new_list = self.left_tree.append_other_standard_names(new_list, refprop_name)

            if self.refprop_name == refprop_name:

                if self.std_name not in initial_list and not self.std_name == refprop_name:

                    new_list.append(self.std_name)

            return new_list


__REFPROP_NAME_TREE = __RefPropNamesTree.initialize_from_xml()


def get_refprop_name(name: str):

    return __REFPROP_NAME_TREE.get_refprop_name(name.lower())


def get_units(name: str, unit_system: str):

    unit_system = unit_system.upper()
    units_dict = __REFPROP_NAME_TREE.get_units_dict(name.lower())

    if units_dict is not None:

        if unit_system in units_dict.keys():

            return units_dict[unit_system]

        else:

            return "Unknown Unit System {}".format(unit_system)

    else:

        return "Unknown property name {}".format(name)


def get_all_refprop_names():

    return __REFPROP_NAME_TREE.append_refprop_names(list())


def get_all_unit_systems():

    units_dict = __REFPROP_NAME_TREE.get_units_dict("p")

    if units_dict is not None:

        return units_dict.keys()

    return None


def get_other_standard_names(refprop_name):

    return __REFPROP_NAME_TREE.append_other_standard_names(list(), refprop_name)




from setuptools import setup
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(

    name='REFPROP_connector',
    version='0.0.5',
    license='GNU GPLv3',

    author='Pietro Ungar',
    author_email='pietro.ungar@unifi.it',

    description='Tools for launching REFPROP calculation and retrieving results from python',
    long_description=long_description,
    long_description_content_type='text/markdown',

    url='https://www.dief.unifi.it/vp-177-serg-group-english-version.html',
    download_url='https://github.com/SERGGroup/REFPROPConnector/archive/refs/tags/0.0.5.tar.gz',

    project_urls={

        'Source': 'https://github.com/SERGGroup/REFPROPConnector',
        'Tracker': 'https://github.com/SERGGroup/REFPROPConnector/issues',

    },

    packages=[

        'REFPROPConnector'

    ],

    install_requires=[

        'pyrebase4',
        'ctREFPROP',
        'setuptools',
        'requests',
        'future'

    ],

    classifiers=[

        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

      ]

)



from REFPROPConnector import ThermodynamicPoint, AbstractThermodynamicPoint, RefPropHandler
import unittest


class TestREFPROPConnector(unittest.TestCase):

    def test_init(self):

        ThermodynamicPoint(["air"], ["1"])
        self.assertEqual(True, True)

    def test_list_properties(self):

        tp = ThermodynamicPoint(["air"], ["1"])
        tp.list_properties()
        self.assertEqual(True, True)

    def test_list_units(self):

        tp = ThermodynamicPoint(["air"], ["1"], unit_system="MASS BASE SI")
        tp.list_unit_systems()
        self.assertEqual(True, True)


if __name__ == '__main__':
    unittest.main()
